/*
 * Window Manager/Screentop for Nano-X
 *
 * Copyright (c) 2000 Century Software <embedded.centurysoftware.com>
 *
 * Written by Jason Kingan and Greg Haerr
 */

/* Battery monitor for those apps with APM support */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#define MWINCLUDECOLORS
#include "nano-X.h"
#include "nanowm.h"
#include "apps.h"

#define APPICON "battery.gif"

/* This is how many times a second the timer refreshes */
#define TIMERRATE 2
#define TTL_TIMEOUT 2

#define TTL_WIN 1
#define BAT_WIN 2

struct battery_t
{
  unsigned short percent;
  unsigned short ttl;
}; 

static struct battery_t global_battery;

/* Some local globals */

static GR_WINDOW_ID ttl_wid = 0;
static unsigned char ttl_timer = 0;
static GR_IMAGE_ID batimage = 0;
static unsigned long timer = 0;
static int apm_fd;

static 
void ttl_exposure(win *window, GR_EVENT_EXPOSURE *ep)
{
  char buf[20];
  GR_GC_ID gc = GrNewGC();
  GR_FONT_ID fontid = GrCreateFont(GR_FONT_GUI_VAR, 0, NULL);

  /* Now tell the user the time to live */
  sprintf(buf, "%d:%2d left\n", global_battery.ttl / 60, 
	  global_battery.ttl % 60);

  GrSetGCForeground(gc, BLACK);
  GrSetGCUseBackground(gc, GR_FALSE);
  GrSetGCFont(gc, fontid);
  GrText(ep->wid, gc, 1, 1, buf, -1, GR_TFTOP);

  GrDestroyFont(fontid);
  GrDestroyGC(gc);
}

static 
void battery_exposure(win *window, GR_EVENT_EXPOSURE *ep)
{
  GR_REGION_ID region = GrNewRegion();
  GR_GC_ID batgc = GrNewGC();
  GR_RECT rect;

  int perc = global_battery.percent;
  int height = 0;

  GrSetGCBackground(batgc, COLOR_TASKBAR);
  if (!batimage)
    {
      char path[256];
      
      sprintf(path, "%s/%s", 
	      get_desktop_config(DESKTOP_IMAGEDIR), APPICON);
      batimage = GrLoadImageFromFile(path, 0);    
    }

  /* Draw the battery */
  GrDrawImageToFit(ep->wid, batgc, 0, 0, -1, -1, batimage);

  /* This will clip our rectangle to the battery */
  /* the good news is that it facilitates drawing the */
  /* actually monitor rectangle, the bad news is that */
  /* we have to recompile everytime the graphic is changed */

  rect.x = 1;
  rect.y = 3;
  rect.width = 6;
  rect.height = 10;

  GrUnionRectWithRegion(region, &rect);
  
  rect.x = 3;
  rect.y = 1;
  rect.width = 2;
  rect.height = 2;
  
  GrUnionRectWithRegion(region, &rect);

  GrSetGCRegion(batgc, region);

  if (perc > 100)
    perc = 100;

  if (perc < 0)
    perc = 0;

  height = (12 * perc) / 100;

  if (perc > 50)
    GrSetGCForeground(batgc, GREEN);
  else if (perc <= 50 && perc > 10)
    GrSetGCForeground(batgc, YELLOW);
  else
    GrSetGCForeground(batgc, RED);

  GrFillRect(ep->wid, batgc, 1, 13 - height, 6, height);

  if (perc < 100)
    {
      GrSetGCForeground(batgc, BLACK);
      GrLine(ep->wid, batgc, 1, 13 - height, 6, 13 - height);
    }

  GrDestroyGC(batgc);
  GrDestroyRegion(region);
}

int get_battery_info(struct battery_t *bptr)
{
  char buf[BUFSIZ];
  char *c;
  int i;

  if (apm_fd < 0) {
	  bptr->percent = 100;
	  bptr->ttl = 120;
	  return 1;
  }

  /* Read the line */
  lseek((int) apm_fd, 0, SEEK_SET);
  read((int) apm_fd, buf, BUFSIZ - 1);

  /* Skip six words in */
  c = buf;

  for(i = 0; i < 6; i++)
    {
      for(;*c != ' ' && *c != 0; c++);
      c++;
    }

  bptr->percent = (unsigned short) atoi(c);

  /* Skip to the next space */
  for(;*c != ' ' && *c != 0; c++);
  c++;

  /* Get the time to live */
  bptr->ttl = (unsigned short) atoi(c);
 
  return(1);
}

static void
redraw_window(win *window, int type)
{
  GR_EVENT_EXPOSURE 	expose;
  GR_WINDOW_INFO		info;
  
  /* dummy full area expose event*/
  GrGetWindowInfo(window->wid, &info);
  expose.type = GR_EVENT_TYPE_EXPOSURE;
  expose.wid = window->wid;
  expose.x = 0;
  expose.y = 0;
  expose.width = info.width;
  expose.height = info.height;
  
  if (type == TTL_WIN)
    ttl_exposure(window, &expose);
  else
    battery_exposure(window, &expose);
}

static void battery_wndproc(win *window, GR_EVENT *ep)
{      
  GR_EVENT_MOUSE *em;

  switch(ep->type)
    {
    case GR_EVENT_TYPE_EXPOSURE:
      
      if (ep->exposure.wid == ttl_wid && ttl_timer)
	ttl_exposure(window, &ep->exposure);
      else
	battery_exposure(window, &ep->exposure);

      break;

    case GR_EVENT_TYPE_BUTTON_DOWN:
      
      if (ttl_timer)
	{
	  ttl_timer = 0;
	  GrUnmapWindow(ttl_wid);
	}
      else
	{
	  ttl_timer = 1;
	  
	  get_battery_info(&global_battery);
	  em = (GR_EVENT_MOUSE *)ep;
	  GrMoveWindow(ttl_wid, em->rootx, em->rooty-15);
	  GrRaiseWindow(ttl_wid);
	  GrMapWindow(ttl_wid);
	}
      break;


    case GR_EVENT_TYPE_TIMEOUT:
      if (ttl_timer)
	{	
	  ttl_timer++;
	  /* Hide the ttl timer after a short timeout */
	  if (ttl_timer == (TTL_TIMEOUT * TIMERRATE))
	    {
	      ttl_timer = 0;
	      GrUnmapWindow(ttl_wid);
	    }
	}
       
      if (timer++ % (TIMERRATE * 30))
	{
	  unsigned short pperc = global_battery.percent;

	  if (!get_battery_info(&global_battery))
	    break;
	  
	  /* Only redraw if the value has changed */
	  if (pperc != global_battery.percent)
	    redraw_window(window, BAT_WIN);
	}
      
      break;
    }
}

void battery_create(void)
{ 
  win *window, *ttlwindow;
  
  GR_WINDOW_ID	  wid;
  GR_SCREEN_INFO  si;
  GR_WM_PROPERTIES props;

  /* Open up a link to the apm data file */
  apm_fd = open("/proc/apm", O_RDONLY);

  if (apm_fd == -1)
    {
      //fprintf(stderr, "Couldn't open /proc/apm\n");
      //return;
    }

  if (!get_battery_info(&global_battery))
    return;
  
  GrGetScreenInfo(&si);
  wid = GrNewWindowEx(GR_WM_PROPS_NODECORATE, 
		      NULL, GR_ROOT_WINDOW_ID,
		      22, si.ws_height + 4, 
		      8, 14, COLOR_TASKBAR);

  window = add_window(wid, GR_ROOT_WINDOW_ID, 0, battery_wndproc);

  ttl_wid = GrNewWindowEx(GR_WM_PROPS_NODECORATE, NULL, GR_ROOT_WINDOW_ID, 22, 
			 si.ws_height - 17, 40, 15, COLOR_TASKBAR);
  props.flags = GR_WM_FLAGS_BORDERSIZE;
  props.bordersize = 1;
  GrSetWMProperties(ttl_wid, &props);

  ttlwindow = add_window(ttl_wid, GR_ROOT_WINDOW_ID, 0, battery_wndproc);

  
  GrSelectEvents(wid, GR_EVENT_MASK_EXPOSURE | GR_EVENT_MASK_TIMEOUT | 
		 GR_EVENT_MASK_BUTTON_DOWN | GR_EVENT_MASK_BUTTON_UP);

  GrSelectEvents(ttl_wid, GR_EVENT_MASK_EXPOSURE);
  GrMapWindow(wid);

  timeout_setproc(battery_wndproc, window);
}
