/*
 * Window Manager/Screentop for Nano-X
 *
 * Copyright (c) 2000 Century Software <embedded.centurysoftware.com>
 *
 * Written by Jason Kingan and Greg Haerr
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "apps.h"
#include "nanowm.h"
#include "catagories.h"
#include "settings.h"
#include "sys_menu.h"

/* This is the list of available catagories */
static NXLISTHEAD catagory_list;

static int CATAGORY_INDEX = 0;

/* This is the currently active catagory */
static CATAGORY *current_catagory = 0;

/* Static function prototypes */
static void string_tolower(int len, char *str);


/* add_catagory()
   Add a new catagory with the specified name
*/

static void format_catagory_title(char *in, char *out)
{
  int i;

  /* Capitalize first letter */
  out[0] = (in[0] & ~0x20);
  
  /* Lower case all others */
  for(i = 1; i < strlen(in); i++)
    out[i] = (in[i] | 0x20);
}

static void add_catagory(char *value)
{ 
  PNXLIST p;

  CATAGORY *cat = nxItemNew(CATAGORY);
  
  if (!cat) return;
  
  strcpy(cat->name, value);
  /* FIXME:  Add further fun stuff here like icons and such */

#ifdef FORMAT_CATAGORY
  /* Format the title */
  format_catagory_title(cat->name, cat->title);
#else
  strcpy(cat->title, cat->name);
#endif /* FORMAT_CATAGORY */


  /* This will be the virtual window that this catagory is associated with */
  cat->index = CATAGORY_INDEX++;

  nxListAdd(&catagory_list, &cat->next);

#ifdef NOTUSED /* Do this later maybe */

  /* Now, we want to insert it alphabetically into the list */

  if (!catagory_list.head)
     {
       nxListAdd(&catagory_list, &cat->next);
       return;
     }
  else
    {
      for(p = catagory_list.head; p; p = p->next)
	{
	  CATAGORY *cur = nxItemAddr(p, CATAGORY, next);

	  if (strcmp(cur->title, cat->title) > 0)
	    {
	      
	      nxListInsertAfter(&catagory_list, &cat->next, p->prev);
	      return;
	    }
	}

      if (!p)
	nxListAdd(&catagory_list, &cat->next);
    }
#endif
}


/* catagory_cb()

   This is the callback for the IniEnumKeyValues call when the catagory information 
   is loaded from the config file 
*/

static void catagory_cb(char *key, char *value, int data)
{
  if (!strcmp(key, "catagory"))
    add_catagory(value);
}

/******* PUBLIC ROUTINES ********/

inline int get_catagory_count()
{
  return(CATAGORY_INDEX);
}

inline int get_current_virtwin()
{
  if (current_catagory)
    return(current_catagory->index);
  return(0);
}

/* get_current_catagory()
   This returns a pointer to the current catagory 
*/

inline CATAGORY* get_current_catagory(void)
{
  return(current_catagory);
}

/* get_cat_by_window()
   
   Given a virtual window, return the catagory associated 
   with that window
*/

CATAGORY* get_cat_by_window(int index)
{
  PNXLIST p = catagory_list.head;
  
  for(;p; p = p->next)
    {
      CATAGORY *cat = nxItemAddr(p, CATAGORY, next);
      
      if (cat->index == index)
	return(cat);
    }

  return(0);
}

CATAGORY *get_next_catagory(CATAGORY *cat)
{
  PNXLIST p = catagory_list.head;
  
  for(;p; p = p->next)
    {
      CATAGORY *nextcat = nxItemAddr(p, CATAGORY, next);

      if (nextcat == cat)
	{
	  if (p->next)
	    return(nxItemAddr(p->next, CATAGORY, next));
	  else
	    {
	      if (catagory_list.head)
		return(nxItemAddr(catagory_list.head, CATAGORY, next));
	    }
	  break;
	}
    }
	 
  return(0);
}

CATAGORY *get_previous_catagory(CATAGORY *cat)
{
  /* Get the previous catagory in the list*/
  PNXLIST p = catagory_list.head;

  for(;p; p = p->next)
    {
      CATAGORY *nextcat = nxItemAddr(p, CATAGORY, next);

      if (nextcat == cat)
	{
	  if (p->prev)
	    return(nxItemAddr(p->prev, CATAGORY, next));
	  else
	    {
	      if (catagory_list.tail)
		return(nxItemAddr(catagory_list.tail, CATAGORY, next));
	    }

	  break;
	}
    }
  
  return(0);
}

/* set_current_catagory()
   This sets the current catagory to the specified index.
   Note:  This does not switch the icons lists or virtual windows.
   use switch_catagory() for that.
*/

int set_current_catagory(int index)
{
  PNXLIST p = catagory_list.head;

  for(;p; p = p->next)
    {
      CATAGORY *cat = nxItemAddr(p, CATAGORY, next);

      if (cat->index == index)
	{
	  current_catagory = cat;
	  return(0);
	}
    }

  return(-1);
}

/* get_catagory_index()
   Given a name, return the index associated with that catagory
*/

int get_catagory_index(char *name)
{
  PNXLIST p = catagory_list.head;
  char			*dupname,								/* For strdup() of name */
  				*dupcat;								/* For strdup() of cat->name */
  
  for(;p; p = p->next)
    {
      CATAGORY *cat = nxItemAddr(p, CATAGORY, next);

		if ( (dupname = strdup(name)) != NULL &&
			 (dupcat = strdup(cat->name)) != NULL)
		{
			/* Make all characters lower case */
			string_tolower(strlen(dupname), dupname);
			string_tolower(strlen(dupcat), dupcat);
			if (!strcmp(dupname, dupcat))
				return(cat->index);
			
			free(dupname);
			free(dupcat);
		} /* end of if */
		else
		{
      		if (!strcmp(name, cat->name))
			return(cat->index);
		} /* end of else */
    }
  
  return(-1);
}

/* add_app_to_catagory()
   Add a application to the specified catagory
*/

int add_app_to_catagory(int index, APP *app)
{
  CATAGORY *cat = get_cat_by_window(index);
  CATAPPLIST *catlist = nxItemNew(CATAPPLIST);

  if (!cat) return(-1);
  if (!catlist) return(-1);

  catlist->app = app;
  nxListAdd(&cat->applist, &catlist->link);
  return(0);
}

/* switch_catagory() 
   
   This switches the current catagory to the one specified by index
   Unlike set_current_catagory, this function switches the icon lists,
   hides windows, and redraws the root window 
*/

int switch_catagory(int index)
{
  CATAGORY *new_cat = get_cat_by_window(index);
  CATAGORY *old_cat = get_current_catagory();

  if (!new_cat) return(-1);
  if (new_cat == old_cat) return(0);

  /* First, hide all of the running apps */
  if (old_cat)
    {
      /* app_hide_windows(old_cat->index); */
      hide_icon_list();
    }

  set_current_catagory(new_cat->index);

  /* Redraw the root window */
  redraw_root_window();

  /* app_show_windows(new_cat->index, 0); */
  
  return(0);
}

/* load_catagory_config()
   Called from main.c, this creates the general catagory, and adds any user catagories 
   from the config file
*/

void load_catagory_config(char *filename)
{
	char		defaultcat[128] = {'\0'},						/* Default catagory from config file */
				defflg = 0;										/* Default catagory flag */
  char buf[4096];

  /* Add the first catagory */

  /*
  ** Added a default catagory config option (that way "Pim" isn't the default) -- 02/14/01 -- J Webb
  */
  CATAGORY_INDEX = 0;
	if (filename)
	{
		if (IniGetString("virtual", "defcatagory", "", defaultcat, sizeof(defaultcat), filename))
		{
			add_catagory(defaultcat);
			defflg=1;
		} /* end of if */
	} /* end of if */

	/* If default value wasn't set, use "Pim" as the default value */
	if (defflg == 0)
		add_catagory("Pim");
  
  /* Parse the config file looking for more catagories */

  if (filename)
    {
      /* Version 1 they were in virtual */
      if (IniGetString("virtual", NULL, "", buf, sizeof(buf), filename))
	IniEnumKeyValues(buf, catagory_cb, 0);

      else if (IniGetString("catagories", NULL, "", buf, sizeof(buf), filename))
	IniEnumKeyValues(buf, catagory_cb, 0);
    }
}

void catagory_free_memory()
{
  PNXLIST p = catagory_list.head;
  
  for(;p; p = p->next)
    {
      CATAGORY *cat = nxItemAddr(p, CATAGORY, next);
      
      if (cat)
	{
	  PNXLIST a = cat->applist.head;
	  
	  for(;a; a = a->next)
	    {
	      CATAPPLIST *catlist = nxItemAddr(a, CATAPPLIST, link);
	      nxListRemove(&cat->applist, &catlist->link);
	    }
	  
	  nxListRemove(&catagory_list, &cat->next);
	}
    }
}

static void
string_tolower(int len, char *str)
{
	char					*ptr;									/* Pointer to traverse string */

	/* Sanity check */
	if (len > strlen(str))
		return;

	for (ptr = str; *ptr && (str - ptr) < len; ptr++)
	{
		if (isalpha(*ptr))
			*ptr = *ptr | 0x20;
	} /* end of for */

	return;
} /* end of string_tolower() */

#ifdef SYSTEM_MENUS
wm_menu_t *create_catagory_menu(int index, char *name)
{
  wm_menu_t *ptr;

  int count = 0;
  PNXLIST p;

  CATAGORY *cat = get_cat_by_window(index);

  if (!cat) return(0);
  
  /* Count how many apps we have in this catagory */
  p = cat->applist.head;

  while(p)
    {
      count++;
      p = p->next;
    }

  if (!count)
    return(0);

  count++;

  ptr = create_system_menu(count);
  if (!ptr) return(0);

  p = cat->applist.head;
  count = 0;

  while(p)
    {
      CATAPPLIST *catlist = nxItemAddr(p, CATAPPLIST, link);      
      APP *app = catlist->app;

      if (!(app->m_flags & FL_NOICON))
	ADD_SYS_MENU_APP(&ptr[count++], app->m_icontitle, app->m_exec, app->m_icon_iid);
      
      p = p->next;
    }

   ADD_SYS_MENU_END(&ptr[count++]); 
   strcpy(name, cat->title);

   return(ptr);
}
#endif
