/* This is where we maintain all of the settings for */
/* The various portions of the window manager        */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nano-X.h"
#include "apps.h"
#include "nanowm.h"
#include "settings.h"

#ifdef LOCK_SETTINGS
/*
**	Certain settings have been "locked" to avoid customization
**	(i.e. removing the Century Logo from the scrtop)
**
*/
#include "hardcodedimage.h"
static char dgifpath[512];										/* Default GIF Path */
static void CreateDefaultGif();
#endif /* end of LOCK_SETTINGS */

/* These is a static structure that holds all */
/* the current values */

static struct
{
  struct
  {
    GR_IMAGE_ID iid;
    int         flags;
  } wallpaper;
  
  GR_COLOR bgcolor;
  GR_COLOR icncolor;
  GR_COLOR icnTxtColor;
} wm_settings;

#define DESKTOP_PATH_COUNT 4

static struct
{
  char *key;
  char *extension;
  char value[256];
} desktop_paths[] = {
  {"prefix", ""},           /* DESKTOP_PREFIX */
  {"imagedir", "images"},   /* DESKTOP_IMAGEDIR */
  {"nxappdir", "bin"},      /* DESKTOP_NXAPPDIR */
  {"sounddir", "sounds"}    /* DESKTOP_SOUNDSDIR */
};

/* This function returns the current image id of the wallpaper */

int wm_get_wallpaper(GR_IMAGE_ID *iid, int *flags)
{
  if (!wm_settings.wallpaper.iid || wm_settings.wallpaper.flags == WALLPAPER_NONE)
    return(0);

  *iid = wm_settings.wallpaper.iid;
  *flags = wm_settings.wallpaper.flags;
  return(1);
}

GR_COLOR wm_get_color(int type)
{
  switch(type)
    {
    case WM_BGCOLOR:
      return(wm_settings.bgcolor);

    case WM_ICONCOLOR:
      return(wm_settings.icncolor);

    case WM_ICONTEXT:
      return(wm_settings.icnTxtColor);
      break;
    }

  return(0);
}

static void wm_set_wallpaper_name(char *name)
{
	char					path[512];								/* Path */

#ifdef LOCK_SETTINGS
  	/* The settings have been locked, the Century Logo will be displayed */
	CreateDefaultGif();
	name=dgifpath;
#endif /* LOCK_SETTINGS */

  if (wm_settings.wallpaper.iid)
    GrFreeImage(wm_settings.wallpaper.iid);

	if (name == NULL || *name == '\0')
	{
		wm_settings.wallpaper.flags = WALLPAPER_NONE;
		return;
	} /* end of if */

  /* First try just to load the direct path */
  wm_settings.wallpaper.iid = GrLoadImageFromFile(name, 0);
  
  if (!wm_settings.wallpaper.iid)
    {
      char *p;
      
      /* cut down the string to the base name */
      if ( (p = strrchr(name, '/')) )
		  p++;
	  else
		  p = name;

	  sprintf(path, "%s/%s", 
		  get_desktop_config(DESKTOP_IMAGEDIR), p);
	  wm_settings.wallpaper.iid = GrLoadImageFromFile(path, 0);
    }

  /* Set the default flags here, these may be changed later */
  if (wm_settings.wallpaper.iid)
    wm_settings.wallpaper.flags = WALLPAPER_CENTERED;
  else
    wm_settings.wallpaper.flags = WALLPAPER_NONE;

#ifdef LOCK_SETTINGS
	/* Now that the image has been loaded, remove the default image */
	unlink(dgifpath);
#endif /* LOCK_SETTINGS */
}

static void wm_set_wallpaper_style(char *value)
{
  char str[20];
  int c;
  int style;

  /* Parse the incoming style */
  /* First job, drive all the letters to lowercase */
  
  strcpy(str, value);

  for(c = 0; c < strlen(str); c++)
    if (str[c] >=65 && str[c] <= 90) str[c] += 32;
  
  style = WALLPAPER_CENTERED;

  if (!strcmp(str, "bottom"))
    style = WALLPAPER_BOTTOM;
  else if (!strcmp(str, "tiled"))
    style = WALLPAPER_TILED;
  else if (!strcmp(str, "fullscreen"))
    style = WALLPAPER_FULLSCREEN;
  else if (!strcmp(str, "none"))
    {
      /* Nothing requested, free anything that was there before */
#ifndef LOCK_SETTINGS
      if (wm_settings.wallpaper.iid)
	GrFreeImage(wm_settings.wallpaper.iid);
  
      wm_settings.wallpaper.iid = 0;
      style = WALLPAPER_NONE;
#else
	  style = WALLPAPER_BOTTOM;
#endif /* ! LOCK_SETTINGS */
    }

  wm_settings.wallpaper.flags = style;
}

static GR_COLOR wm_parse_color(char *value)
{
  unsigned char r, g, b;
  int colorval;
  
  if (value[0] != '#')
    return(0);
  
  colorval = strtol(value + 1, NULL, 16);
  r = (colorval >> 16) & 0xFF;
  g = (colorval >> 8) & 0xFF;
  b = (colorval & 0xFF);

  return(GR_RGB(r, g, b));
}
 
static void wm_set_color(int type, GR_COLOR color)
{
  switch(type)
    {
    case WM_BGCOLOR:
      wm_settings.bgcolor = color;
      break;

    case WM_ICONCOLOR:
      wm_settings.icncolor = color;
      break;

    case WM_ICONTEXT:
      wm_settings.icnTxtColor = color;
      break;
    }
}

static void settings_cb(char *key, char *value, int data)
{
  if (!strcmp(key, "bgImage"))
    {
      wm_set_wallpaper_name(value);
      return;
    }

  if (!strcmp(key, "bgImageStyle"))
    {
      wm_set_wallpaper_style(value);
      return;
    }

  if (!strcmp(key, "bgColor"))
    {
      wm_set_color(WM_BGCOLOR, wm_parse_color(value));
      return;
    }

  if (!strcmp(key, "iconTxtForeground"))
    {
      wm_set_color(WM_ICONTEXT, wm_parse_color(value));
      return;
    }

  if (!strcmp(key, "iconTxtBackground"))
    {
      wm_set_color(WM_ICONCOLOR, wm_parse_color(value));
      return;
    }
}

void wm_load_config_settings()
{
  char *path = nxGetConfigFile(NX_CONFIG_MAIN);
  char buf[4096];

  wm_set_defaults();
  
  if (path)
    {
      if (IniGetString("wm", NULL, "", buf, sizeof(buf), path))
	IniEnumKeyValues(buf, settings_cb, 0);
    }
}

void wm_set_defaults()
{
  char path[512];
  
  sprintf(path, "%s/%s", 
	  get_desktop_config(DESKTOP_IMAGEDIR), DEFAULT_WALLPAPER);

  wm_set_wallpaper_name(path);
  wm_set_wallpaper_style("bottom");

  wm_set_color(WM_BGCOLOR, DEFAULT_BGCOLOR);
  wm_set_color(WM_ICONCOLOR, DEFAULT_ICONCOLOR);
  wm_set_color(WM_ICONTEXT, DEFAULT_ICONTEXT);
}

char* get_desktop_config(int flag)
{
  if (flag >= DESKTOP_PATH_COUNT)
    return(0);

  if (strlen(desktop_paths[flag].value))
    return(desktop_paths[flag].value);
  else
    return(0);
}

/* Go through and expand out any variables in the path name */

void desktop_expand_path(char *input, char *output, int outsize)
{
  str_replace_variable("$prefix", get_desktop_config(DESKTOP_PREFIX),
		       input, output, outsize);
}
    
void load_desktop_paths(char *filename)
{
  int i, j;

  char buf[4096];

  for(i = 0; i < DESKTOP_PATH_COUNT; i++)
    bzero(desktop_paths[i].value, sizeof(desktop_paths[i].value));
  
  /* Get the paths from the config file */

  if (filename)
    {
      if (IniGetString("desktop", NULL, "", buf, sizeof(buf), filename))

	for(i = 0; i < DESKTOP_PATH_COUNT; i++)
	  {
	    IniGetString("desktop", desktop_paths[i].key, "", desktop_paths[i].value, 
			 sizeof(desktop_paths[i].value), filename);

	    /* Strip the white space at the end */

	    if ((j = strlen(desktop_paths[i].value) - 1))
	      {
		for(; desktop_paths[i].value[j] == ' '; j--)
		  desktop_paths[i].value[j] = 0;
	      }
	  }
    }    
      /* Check for defaults, and expand any relative paths */

  
  for(i = 0; i < DESKTOP_PATH_COUNT; i++)
    {
      if (i == DESKTOP_PREFIX)
	{
	  if (!strlen(desktop_paths[DESKTOP_PREFIX].value))
	    sprintf(desktop_paths[DESKTOP_PREFIX].value, ".");
	  else
	    {
	      if (desktop_paths[DESKTOP_PREFIX].value[0] != '/')
		{
		  char buffer[strlen(desktop_paths[DESKTOP_PREFIX].value)];

		  /* Its a relative path, add the current working directory on the */
		  /* front */

		  strcpy(buffer, desktop_paths[DESKTOP_PREFIX].value);
		  getcwd(desktop_paths[DESKTOP_PREFIX].value, sizeof(desktop_paths[DESKTOP_PREFIX].value));
		  strcat(desktop_paths[DESKTOP_PREFIX].value, "/");
		  strcat(desktop_paths[DESKTOP_PREFIX].value, buffer);
		}
	    }

	  if (desktop_paths[DESKTOP_PREFIX].value[strlen(desktop_paths[DESKTOP_PREFIX].value) - 1] == '/')
	    desktop_paths[DESKTOP_PREFIX].value[strlen(desktop_paths[DESKTOP_PREFIX].value) - 1] = 0;
	}
      else
	{
	  if (!strlen(desktop_paths[i].value))
	    sprintf(desktop_paths[i].value, "%s/%s", 
		    desktop_paths[DESKTOP_PREFIX].value,
		    desktop_paths[i].extension);
	}
    }
}

      
#ifdef LOCK_SETTINGS
static void
CreateDefaultGif(void)
{
	char					outPath[255] = {'\0'};					/* Path to outstream */
	int						ii,										/* Array index */
							len;									/* Length */
	FILE					*outstrm;								/* Outstream file */

	/* Set outpath */
	for (len = ii = 0; ii < DESKTOP_PATH_COUNT; ii++)
	{
		if (!strcmp(desktop_paths[ii].key, "imagedir"))
		{
			len = sprintf(outPath, "%s/", desktop_paths[ii].value);
			break;
		} /* end of if */
	} /* end of for */

	sprintf(outPath + len, "%s", DEFAULT_WALLPAPER);

	if ( (outstrm = fopen(outPath, "wb")) == NULL)
	{
		if (errno == EPERM || errno == EACCES)
		{
			/* Don't have the necessary permissions, try the local directory */
			char			*cp;									/* Character pointer */

			/* Nano-X does not like relative paths, so only give it an absolute path */
			if ( (cp = strrchr(outPath, '/')) == NULL)
				cp = outPath;
			else
				cp++;
			len = strlen(getcwd(outPath, sizeof(outPath)));
			sprintf(outPath + len, "/%s", DEFAULT_WALLPAPER);
			if ( (outstrm = fopen(outPath, "wb")) == NULL)
				return;
		} /* end of if */
		else
		{
			return;
		} /* end of else */
	} /* end of if */

	/* Store path in static memory */
	memcpy(dgifpath, outPath, strlen(outPath));
	fwrite(hardcodedimage, sizeof(char), SIZE_hardcodedimage, outstrm);
	fclose(outstrm);
	return;
} /* end of CreateDefaultGif() */
#endif /* LOCK_SETTINGS */

