/* SOUND_DSP.C

   Library:       nxIO - IO Library
   Description:  DSP driver for recording and playing sound
   Author:       Jordan Crouse

   Copyright 2000, Century Software Embedded Technologies
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/soundcard.h>

#include "nxio.h"
#include "sound_dsp.h"

#define AUDIO_FILE "/dev/dsp"

/* This array ties the NXIO_MIXER defines to the */
/* LINUX defines */

static int nxio2linux_values[NXIO_MIXER_COUNT] = { 
  SOUND_MIXER_VOLUME,
  SOUND_MIXER_BASS,
  SOUND_MIXER_TREBLE,
  SOUND_MIXER_MIC,
  SOUND_MIXER_LINE1,
  SOUND_MIXER_LINE2
};

static int OpenAudioDevice(int direction)
{
  int omode;
  int fd;
  
  if (direction == NXIO_PLAY)
    omode = O_WRONLY;
  else
    omode = O_RDONLY | O_NONBLOCK;

  fd = open(AUDIO_FILE, omode, 0);
  
  if (fd == -1)
    perror("OPEN_AUDIO");
  
  return(fd);
}

void ParseWAVHeader(dspWaveHeader *hdr, nxio_audio_t *playback)
{
  playback->stereo = (hdr->modus == WAVE_STEREO) ? 1 : 0;
  playback->sample = hdr->bits_per_sample;
  playback->speed = hdr->spd;
  playback->size = hdr->datalen;
}

int dspOpenStream(int direction, nxio_audio_t *settings)
{
  int err = NXIO_ERROR;
  int tmp, bsize;
  int fd;
  
  fd = OpenAudioDevice(direction);

  if (fd < 0) return(err);

  /* Sync the device */
  if (ioctl(fd, SNDCTL_DSP_SYNC, NULL) < 0)
    {
      perror("OPEN_STREAM (SYNC)");
      goto exit_dspOpenStream;
    }
  
  /* Set the sample size */
  tmp = settings->sample; 

  ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &tmp);

  if (tmp != settings->sample)
    {
      /* Failure - Try setting it to 8 bits */
      tmp = 8;
      ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
      if (tmp != 8)
	{
	  perror("OPEN_STREAM (SIZE)");
	  goto exit_dspOpenStream;
	}
    }

  /* Set the sample rate */
  if (ioctl(fd, SNDCTL_DSP_SPEED, &settings->speed) < 0)
    {
      perror("OPEN_STREAM (SPEED)");
      goto exit_dspOpenStream;
    }  

  /* Finally, set the stereo setting */
  if (settings->stereo)
    if (ioctl (fd, SNDCTL_DSP_STEREO, &settings->stereo) < 0) 
      settings->stereo = 0;
  
  /* Last step, grab the block size and return it to the caller */
  /* Get the biggest block we can read */

  bsize = -1;
  ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &bsize);
  
  if (bsize < 4 || bsize > 65536) 
    {
      if (bsize == -1)
	{
	  perror ("OPEN_STREAM (BLOCKSIZE)");
	  goto exit_dspOpenStream;
	}
      bsize = 1024;
    }
  
  settings->blocksize = bsize;
  err = NXIO_OK;
  return(fd);

 exit_dspOpenStream:
  if (fd >= 0)
    close(fd);

  return(err);
}

int dspStreamRecord(int fd, unsigned char *buffer, int size)
{
  int len = read(fd, buffer, size);
  
  if (len < 0)
    {
      if (errno == EAGAIN)
	return(NXIO_AGAIN);
      
      perror("dspStreamRecord");
      return(NXIO_ERROR);
    }

  return(len);
}

int dspStreamPlay(int fd, unsigned char *buffer, int size)
{
  int len = write(fd, buffer, size);
  
  if (len < 0)
    {
      if (errno == EAGAIN)
	return(NXIO_AGAIN);

      perror("PLAY_STREAM");
      return(NXIO_ERROR);
    }

  return(len);
}

int dspGetWAVFileStats(char *filename, nxio_audio_t *settings)
{
  dspWaveHeader wavheader;
  int infd;

  if ((infd = open(filename, O_RDONLY, 0)) == -1)
    {
      perror("OPEN_WAV");
      return(NXIO_ERROR);
    }
  
  /* Read and parse the header */
  if (read(infd, (char *) &wavheader, sizeof(dspWaveHeader)) 
      < sizeof(dspWaveHeader))
    {
      perror("READ_WAV");
      close(infd);
      return(NXIO_ERROR);
    }
  
  ParseWAVHeader(&wavheader, settings); 
  close(infd);
  return(NXIO_OK);
}

#define RIFF            0x46464952
#define WAVE            0x45564157
#define FMT             0x20746D66
#define DATA            0x61746164
#define PCM_CODE        1

int dspWriteWAVFile(char *filename, nxio_audio_t *settings, unsigned char *buffer)
{
  int fd;
  dspWaveHeader hdr;
  
  hdr.main_chunk = RIFF;
  hdr.length     = settings->size + sizeof(dspWaveHeader) - 8;
  hdr.chunk_type = WAVE;
  hdr.sub_chunk  = FMT;
  hdr.sc_len     = 16;
  hdr.format     = PCM_CODE;
  hdr.modus      = settings->stereo ? 2 : 1;
  hdr.spd         = settings->speed;
  hdr.bytes_per_sample = ((settings->sample == 8) ? 1 : 2) * (settings->stereo ? 2 : 1);
  hdr.bytes_per_sec = settings->speed * hdr.modus * hdr.bytes_per_sample;
  hdr.bits_per_sample  = settings->sample;
  hdr.data_chunk = DATA;
  hdr.datalen= settings->size;

  if ((fd = open(filename, O_WRONLY | O_CREAT, S_IREAD | S_IWRITE)) == -1)
    {
      perror("OPEN WAV");
      return(NXIO_ERROR);
    }
  
  write (fd, &hdr, sizeof(dspWaveHeader));
  write (fd, buffer, settings->size);
  close(fd);
  return(NXIO_OK);
}

int dspReadWAVFile(char *filename, nxio_audio_t *settings, unsigned char *buffer, int maxsize)
{
  int infd, len;
  dspWaveHeader wavheader;

  if ((infd = open(filename, O_RDONLY, 0)) == -1)
    {
      perror("OPEN WAV");
      return(NXIO_ERROR);
    }
  
  /* Read and parse the header */
  if (read(infd, (char *) &wavheader, sizeof(dspWaveHeader)) 
      < sizeof(dspWaveHeader))
    {
      perror("READ WAV");
      close(infd);
      return(NXIO_ERROR);
    }

  ParseWAVHeader(&wavheader, settings); 

  if (settings->size > maxsize)
    {
      close(infd);
      return(NXIO_FILE_TOO_BIG);
    }

  len = read(infd, buffer, settings->size);
  
  if (len == -1 || len != settings->size)
    {
      if (len == -1)
	perror("READ WAV");

      close(infd);
      return(NXIO_ERROR);
    }
  else
    {
      close(infd);
      return(len);
    }
}

int dspPlayWAVFile(char *filename)
{
  int err = NXIO_ERROR;
  char *audiobuffer;

  dspWaveHeader wavheader;

  int dev = 0;
  int infd = 0;

  nxio_audio_t playback;

  /* Read the WAV file */

  if ((infd = open(filename, O_RDONLY, 0)) == -1)
    {
      perror("OPEN WAV");
      return(NXIO_ERROR);
    }
  
  /* Read and parse the header */
  if (read(infd, (char *) &wavheader, sizeof(dspWaveHeader)) 
      < sizeof(dspWaveHeader))
    {
      perror("READ WAV");
      close(infd);
      return(NXIO_ERROR);
    }

  ParseWAVHeader(&wavheader, &playback); 

  /* Now open up the stream */
  dev = dspOpenStream(NXIO_PLAY, &playback);
  
  if (dev == NXIO_ERROR)
    {
      close(infd);
      return(NXIO_ERROR);
   }

  audiobuffer = (char *) malloc(playback.blocksize);

  while(1)
    {
      int inlen;
      
      /* Read a block in from the file */
      inlen = read(infd, (char *) audiobuffer, playback.blocksize);
      
      if (inlen == -1)
	{
	  perror("READ_WAV");
	  break;
	}
      
      if (inlen == 0) 
	{
	  err = NXIO_OK; /* No error */
	  break;
	}

      if (write (dev, (char *) audiobuffer, inlen) != inlen)
	{
	  perror("WRITE_AUDIO");
	  break;
	}

      if (inlen < playback.blocksize) 
	{
	  err = NXIO_OK; /* No error */
	  break;
	}
    }

  close(dev);
  close(infd);
  return(err);

  return(err);
}

static int open_mixer()
{
  int fd;

  fd = open("/dev/mixer", O_RDWR);

  if (fd == -1)
    perror("OPEN_MIXER");
  
  return(fd);
}

static unsigned long read_mixer(int device)
{
  int mixer = 0;
  int level = -1;

  if ((mixer = open_mixer()) == -1)
    return(-1);

  if (ioctl(mixer, MIXER_READ(device), &level) == -1)
    {
      perror("MIXER_READ (IOCTL)");
      level = -1;
    }

  close(mixer);
  return(level);
}

static int write_mixer(int device, unsigned long level)
{
  int err = -1;
  int mixer = 0;

  if ((mixer = open_mixer()) == -1)
    return(-1);

  if (ioctl(mixer, MIXER_WRITE(device), &level) == -1)
    {
      perror("MIXER_WRITE (IOCTL)");
      err = -1;
    }
  else
    err = 0;

  close(mixer);
  return(err);
}

int mixer_get_devices(unsigned long *bitmask)
{
  int i, j;
  unsigned long readin;

  if ((readin = read_mixer(SOUND_MIXER_DEVMASK)) == -1)
    return(NXIO_ERROR);

  *bitmask = 0;

  for(i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    {
      if (readin & (1 << i))
	{	  
	  for(j = 0; j < NXIO_MIXER_COUNT; j++)
	    {
	      if (nxio2linux_values[j] == i)
		*bitmask |= (1 << j);
	    }
	}
    }
  
  return(NXIO_OK);
}

int mixer_set_level(int device, nxio_level_t *level)
{
  int linux_device;
  unsigned long set_level;

  if (device > NXIO_MIXER_COUNT)
    return(NXIO_BAD_DEVICE);
  else
    linux_device = nxio2linux_values[device];

  set_level = ((level->right * 0x7f) / 100) << 8;
  set_level |= (level->left * 0x7f) / 100;

  
  if (write_mixer(linux_device, (unsigned long) set_level) == -1)
    return(NXIO_ERROR);
  else
    return(NXIO_OK);
}

int mixer_get_level(int device, nxio_level_t *level)
{
  int linux_device;
  int levelin;
  int l,r;

  if (device > NXIO_MIXER_COUNT)
    return(NXIO_BAD_DEVICE);
  else
    linux_device = nxio2linux_values[device];

  if ((levelin = read_mixer(linux_device)) == -1)
    return(NXIO_ERROR);
 
  l = levelin & 0x7f;
  r = (levelin >> 8) & 0x7f;

  level->left = l * 100 / 0x7f;
  level->right = r * 100 / 0x7f;

  return(NXIO_OK);
}
  
