#ifndef RSCTOOLS_H
#define RSCTOOLS_H



/*
**
** Imported "Include" Files
**
*/
#include <stdio.h>



/*
**
** Global Enumeration Definitions
**
*/
typedef enum
{
  PARMTYPE_BYTE,
  PARMTYPE_INTEGER,
  PARMTYPE_DOUBLE,
  PARMTYPE_STRING,
  PARMTYPE_LAST
}ParmType;



/*
**
** Global Structure Definitions
**
*/
typedef struct
{
  ParmType type;
  void *result;
}ParmEntry;



/*
**
**  NAME: locateKeyword()
**
** USAGE: int locateKeyword(FILE *rfp, char *keyword);
**
** DESCR: This function will search the specified file for the specified
**        keyword. The search begins at the current file position, and if
**        the keyword is not found when the end of the file is reached,
**        the search will continue at the beginning of the file.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "keyword"
**        parameter points to the buffer that is loaded with the keyword.
**
** RETRN: If the keyword is found, '1' is returned and the file's position
**        will be set to the first character following the keyword. If the
**        keyword is not found, '0' is returned. If an error occurs during
**        function execution, '-1' is returned.
**
*/
int locateKeyword(FILE *, char *);



/*
**
**  NAME: previewNextKeyword()
**
** USAGE: int previewNextKeyword(FILE *rfp, char *result);
**
** DESCR: This function will recover the next keyword entry found in the
**        specified file. The search begins at the current file position,
**        and terminates when the end of the file is reached.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "result"
**        parameter points to a buffer where the next keyword found in the
**        file will be stored.
**
** RETRN: If successful, '1' is returned and the result buffer will be
**        loaded with the next keyword. If no keyword is found, '0' is
**        returned. If an error occurs during function execution, '-1' is
**        returned. In any case, the file's position is not altered.
**
*/
int previewNextKeyword(FILE *, char *);



/*
**
**  NAME: getParmTotal()
**
** USAGE: int getParmTotal(FILE *rfp);
**
** DESCR: This function will extract the number of parameters associated
**        with a keyword. After the keyword has been located, the number
**        of parameters that follow the keyword are calculated.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened.
**
** RETRN: This function returns the number of parameters associated with the
**        keyword. If an error occurs during function execution, '-1' is
**        returned.
**
*/
int getParmTotal(FILE *);



/*
**
**  NAME: getByteParm()
**
** USAGE: int getByteParm(FILE *rfp, char *result, unsigned int count);
**
** DESCR: This function will recover the byte values associated with a
**        keyword. After the keyword is located, the parameters that follow
**        are recovered as signed byte values.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "result"
**        parameter points to an array of bytes where each byte value will
**        be stored. The "count" parameter contains the number of byte
**        parameters that will be recovered.
**
** RETRN: This function returns the number of byte parameters recovered.
**        If an error occurs during function execution, '-1' is returned.
**
** NOTES: If the "count" parameter is zero, then this function will recover
**        any and all byte parameters associated with the keyword. This
**        process will continue until the next keyword is found or the
**        end-of-file is detected.
**
*/
int getByteParm(FILE *, char *, unsigned int);



/*
**
**  NAME: getIntegerParm()
**
** USAGE: int getIntegerParm(FILE *rfp, int *result, unsigned int count);
**
** DESCR: This function will recover the integer values associated with a
**        keyword. After the keyword is located, the parameters that follow
**        are recovered as signed integer values.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "result"
**        parameter points to an array of integers where each integer value
**        will be stored. The "count" parameter contains the number of
**        integer parameters that will be recovered.
**
** RETRN: This function returns the number of integer parameters recovered.
**        If an error occurs during function execution, '-1' is returned.
**
** NOTES: If the "count" parameter is zero, then this function will recover
**        any and all integer parameters associated with the keyword. This
**        process will continue until the next keyword is found or the
**        end-of-file is detected.
**
*/
int getIntegerParm(FILE *, int *, unsigned int);


/*
**
**  NAME: getDoubleParm()
**
** USAGE: int getDoubleParm(FILE *rfp, double *result, unsigned int count);
**
** DESCR: This function will recover the double values associated with a
**        keyword. After the keyword is located, the parameters that follow
**        are recovered as floating-point, "double" values.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "result"
**        parameter points to an array of doubles where each double value
**        will be stored. The "count" parameter contains the number of
**        double parameters that will be recovered.
**
** RETRN: This function returns the number of double parameters recovered.
**        If an error occurs during function execution, '-1' is returned.
**
** NOTES: If the "count" parameter is zero, then this function will recover
**        any and all "double" parameters associated with the keyword. This
**        process will continue until the next keyword is found or the
**        end-of-file is detected.
**
*/
int getDoubleParm(FILE *, double *, unsigned int);



/*
**
**  NAME: getStringParm()
**
** USAGE: intgetStringParm(FILE *rfp, char **result, unsigned int count);
**
** DESCR: This function will recover the text strings associated with a
**        keyword. After the keyword is located, the parameters that follow
**        are recovered as text strings.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "result"
**        parameter points to an array of string pointers. Each string
**        pointer points to a buffer where one of the text strings is to be
**        stored. The "count" parameter contains the number of text strings
**        that will be recovered.
**
** RETRN: This function returns the number of text strings recovered.
**        If an error occurs during function execution, '-1' is returned.
**
** NOTES: If the "count" parameter is zero, then this function will recover
**        any and all text strings associated with the keyword. This process
**        will continue until the next keyword is found or the end-of-file
**        is detected.
**
*/
int getStringParm(FILE *, char **, unsigned int);



/*
**
**  NAME: getMixedParm()
**
** USAGE: int getMixedParm(FILE *rfp, ParmEntry *parmlist, unsigned int count);
**
** DESCR: This function will recover the parameters associated with a
**        keyword. The parameters that are recovered may be any combination
**        of byte, integer, double, or text strings.
**
** PARMS: The "rfp" parameter points to the FILE structure that was returned
**        when the resource file was successfully opened. The "parmlist"
**        parameter points to an array of ParmEntry structures, where each
**        structure is initialized with the expected parameter type and a
**        pointer to the location where the resulting data will be stored.
**        The "count" parameter contains the number of parameters that will
**        be recovered from the resource file.
**
** RETRN: This function returns the number of parameters recovered. If an
**        error occurs during function execution, '-1' is returned.
**
*/
int getMixedParm(FILE *, ParmEntry *, unsigned int);



/*
**
**  NAME: getRawString();
**
** USAGE: void getRawString(char *result);
**
** DESCR: This function will recover the raw text on the line associated
**        with the keyword. The newline on the end of the line will be
**        removed during the recovery.
**
** PARMS: The "result" parameter points to the buffer that will hold the
**        resulting text.
**
** RETRN: Nothing.
**
*/
void getRawString(char *);

#endif   /* RSCTOOLS_H */
