/*******************************************************************************\
**
**	Program:	dvdplayer.cpp
**	Desc:		Cannibalized version of the mpegplayer.cpp for dvd play back.
**				This provides a "wrapper" application for the National Semi Conductor
**				Middleware API, that will be called via the CET wmcmd plugin modules
**				to provide dvd playback from within the WebMedia Applications
**	History:	05/04/01	J Webb <jeffw@censoft.com>
**							Initial Version
**
\*******************************************************************************/

/*
 * Feature test switches
 */


/*
 * System header files
 */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

// National middleware includes
#include <anlg_video.h>
#include <vmwdvdapi.h>

/*
 * Local header files
 */
#include "sharedclient.h"


/*
**
** Local Constant Defintions
**
*/
#define TRUE 1
#define FALSE 0

typedef short dvdflag_t;
// Values for the dvd state flag
// Note: These values may change depending on the implementation of searching, trick play,
// menus, etc
#define		DVD_ACTION_PLAY					0x0001					// DVD is in play mode
#define		DVD_ACTION_PAUSE				0x0002					// DVD is in paused mode
#define		DVD_ACTION_SCAN					0x0004					// DVD is in scan mode
#define		DVD_ACTION_MASK					0x000F					// Mask defining the action
#define		DVD_SCAN_FORWARD				0x0010					// Forward scan
#define		DVD_SCAN_REVERSE				0x0020					// Reverse scan
#define		DVD_SCAN_SLOW					0x0040					// Slow scan speed
#define		DVD_SCAN_FAST					0x0080					// Fast scan speed
#define		DVD_SCAN_MASK					0x00f0					// Mask defining scan attributes
#define		DVD_CLEAR_SCAN					0x00f4					// Quick macro to clear everything associated with a scan
#define		DVD_EXIT						0x8000					// Exit

#define		DVD_SCAN_SLOWSPD				0.1						// Slow speed value
#define		DVD_SCAN_FASTSPD				2.0						// Fast speed value
#define		DVD_SCAN_NORMSPD				1.0						// Normal viewing speed

#define		DVD_SCAN_SPEED					DVD_SCAN_FASTSPD		// Default speed value

/*
**
** Local Variable Declarations
**
*/


/*
**
** Static Function prototypes
**
*/
static void ProcessScan(dvdflag_t *stateflag, int mode);
static void ProcessPlay(dvdflag_t *stateflag);
static void ProcessPause(dvdflag_t *stateflag);
static void ProcessResize(void);
static void ProcessStop(dvdflag_t *stateflag);
static int stopScan(dvdflag_t stateflag);

/*
**
**
**
*/
static int waitForCommand(void)
{
  int cmd;

  while(1)
  {
    usleep(100000);
    if(cmd_block->status != WMSTATUS_START)
      continue;

	// Set the value, in case something else comes along and over-rides this
	// Note:  The above shouldn't happen, but just in case...
	cmd = cmd_block->command;
	if ( (cmd == WMCMD_DVD_FWD) ||
		 (cmd == WMCMD_DVD_PAUSE) ||
		 (cmd == WMCMD_DVD_PLAY) ||
		 (cmd == WMCMD_DVD_REV) ||
		 (cmd == WMCMD_DVD_STOP) )
	{
		break;
	} // end of if 
  }
  cmd_block->status = WMSTATUS_DONE;
  while(cmd_block->status != WMSTATUS_IDLE)
    ;

  return(cmd);
}


/*******************************************************************************\
**
**	Function:	static void ProcessPlay()
**	Desc:		Handles the play user selection
**	Accepts:	dvdflag_t *stateflag = Ptr to the DVD state flag
**	Returns:	Nothing (void)
**
\*******************************************************************************/
static void
ProcessPlay(dvdflag_t *stateflag)
{
	int					rc;											// Result code

	// Note:	Changed logic in response to NSC Validation Issue 106.74.
	// 			When play button is pressed, it will now return to playback mode
	// 			regardless of what state it is in.  06/06/01 jmw

	if (!((*stateflag) & DVD_ACTION_MASK))
	{
		if ( (rc = VMW_DVD_TitlePlay(1)) == VMW_OK)
		{
			*stateflag |= DVD_ACTION_PLAY;
		}
	} // end of if 
	else
	{
		if ((*stateflag) & DVD_ACTION_PAUSE)
		{
			if ( (rc = VMW_DVD_PauseOff()) == VMW_OK)
				*stateflag &= ~DVD_ACTION_PAUSE;
		} // end of else-if
		if ((*stateflag) & DVD_ACTION_SCAN)
		{
			if ( (rc = stopScan(*stateflag)) == VMW_OK)
				*stateflag &= ~DVD_CLEAR_SCAN;
			fprintf(stdout, "Stopping scan in play, rc=%d, stateflag=0x%x.\n", rc,
				*stateflag);
		} // end of if
	} // end of else

	return;
} // end of ProcessPlay()

/*******************************************************************************\
**
**	Function:	static void ProcessPause()
**	Desc:		Handles the Pause user selection
**	Accepts:	dvdflag_t *stateflag = Ptr to the DVD state flag
**	Return:		Nothing (void)
**
\*******************************************************************************/
static void
ProcessPause(dvdflag_t *stateflag)
{
	if ( (*stateflag) & DVD_ACTION_PLAY)
	{
		int				rc;											// Result code
		if ( (*stateflag) & DVD_ACTION_PAUSE)
		{
			if ( (rc = VMW_DVD_PauseOff()) == VMW_OK)
				*stateflag &= ~DVD_ACTION_PAUSE;
		} // end of if
		else
		{
			if ( (rc = VMW_DVD_PauseOn()) == VMW_OK)
				*stateflag |= DVD_ACTION_PAUSE;
		} // end of else
	} // end of if

	return;
} // end of ProcessPause()

/*******************************************************************************\
**
**	Function:	static void ProcessScan()
**	Desc:		Handles the forward scan user selection
**	Accepts:	dvdflag_t *stateflag = Ptr to the DVD state flag
**				int mode = Type of scan
**				int mode = Mode of scan
**					DVD_SCAN_FORWARD = forward scan
**					DVD_SCAN_REVERSE = reverse scan
**	Returns:	Nothing (void)
**
\*******************************************************************************/
static void
ProcessScan(dvdflag_t *stateflag, int mode)
{
	int					rc;											// Generic result code

	// Four possible scenarios:
	// 1. In idle mode -- do nothing
	// 2. In play mode -- start scan
	// 3. In scan mode same direction -- stop scan
	// 4. In scan mode opposite direction -- reverse scan

	fprintf(stdout, "ProcessScan(): Top, stateflags = 0x%x, mode = %s.\n", *stateflag,
			mode == DVD_SCAN_FORWARD ? "Forward" : "Reverse");

	// We must be in play mode to be able to scan  --- for right now anyway
	if (!((*stateflag) & DVD_ACTION_PLAY))
		return;

	if ( (*stateflag) & DVD_ACTION_SCAN)
	{
		// Stop the current scan because either a stop was requested or,
		// a reverse scan was requested...
		rc = stopScan(*stateflag);
		fprintf(stdout, "Stopped current scan, rc=%d\n", rc);
	} // end of if 
	
	// Determine if we need to start scanning or reverse scanning direction
	if (!((*stateflag & DVD_ACTION_SCAN)) ||
		(*stateflag) & ((mode == DVD_SCAN_FORWARD) ? DVD_SCAN_REVERSE : DVD_SCAN_FORWARD))
	{
		if (mode == DVD_SCAN_FORWARD)
			rc = VMW_DVD_ForwardScan(DVD_SCAN_SPEED,TRUE);
		else
			rc = VMW_DVD_BackwardScan(DVD_SCAN_SPEED,TRUE);
		if (rc == VMW_OK)
		{
			*stateflag &= ~(DVD_CLEAR_SCAN);
			*stateflag |= (DVD_ACTION_SCAN | mode);
			fprintf(stdout, "Started %s scan.\n", mode == DVD_SCAN_FORWARD ? "Forward" : "Reverse");
		} // end of if 
	} // end of if 
	else
	{
		*stateflag &= ~(DVD_CLEAR_SCAN);
	} // end of else

	fprintf(stdout, "ProcessScan():End, stateflags = 0x%x.\n", *stateflag);
	return;
} // end of ProcessScan()

/*******************************************************************************\
**
**	Function:	static void ProcessStop()
**	Desc:		Handles the stop user selection
**	Accepts:	dvdflag_t *stateflag = Ptr to the DVD state flag
**	Return:		Nothing (void)
**
\*******************************************************************************/
static void
ProcessStop(dvdflag_t *stateflag)
{
	int					rc;											// Result code

	// Note:	Changed logic slightly in response to NSC Validation Issue 106.75.
	// 			When stop button is pressed, it will completely stop playback and
	// 			the screen will go black -- thus starting the dvd at the beginning 
	// 			upon future play selections --- 06/06/01 jmw
	// Will call the stop function, but first needs to determine if a scan is in progress
	if ( (*stateflag) & DVD_SCAN_MASK)
	{
		stopScan(*stateflag);
		*stateflag &= ~(DVD_CLEAR_SCAN);
	} // end of if 
	if ( (rc = VMW_DVD_Stop()) == VMW_OK)
	{
		*stateflag = 0;
	} // end of else-if

	return;
} // end of ProcessRevScan()

/*******************************************************************************\
**
**	Function:	static int stopScan()
**	Desc:		Stops the scan in progress
**	Accepts:	dvdflag_t stateflag = DVD State flag values
**	Returns:	int; return value from the API call
**
\*******************************************************************************/
static int
stopScan(dvdflag_t stateflag)
{
	int					rc = -1;									// Result code

	if (stateflag & DVD_ACTION_SCAN)
	{
		if (stateflag & DVD_SCAN_FORWARD)
			rc = VMW_DVD_ForwardScan(0,TRUE);
		else
			rc = VMW_DVD_BackwardScan(0,TRUE);
	} // end of if

	return (rc);
} // end of stopScan()


/*
**
**
**
*/
int main(int argc, char *argv[])
{
	int					dvdrc,										// Result code from dvd api calls
						result,										// Command value
						video;										// Video source
	dvdflag_t			stateflag = 0;								// Current dvd state

	// TODO:
	// 	During first phase (simple dvd playback), I'm just concentrating on
	// 	play, stop, pause, fscan, & rscan

  // recover the shared memory cmd_block pointer
	
printf("dvdplayer:DEBUG In wrapper application\n");

  if(clientMemoryAttach() != 0)
  {
    fprintf(stderr,"Unable to attach to shared memory...\n");
    exit(1);
  }

  // Open the dvd device
fprintf(stdout, "dvdplayer:DEBUG Opening the dvd device\n");
  if ( (dvdrc = VMW_DVD_Open()) != VMW_OK)
  {
	  fprintf(stderr, "VMW_DVD_Open() failed: Error = %d, exiting.\n", dvdrc);
	  exit (EXIT_FAILURE);
  } // end of if 
  ProcessPlay(&stateflag);
	
  // Get the initial command 
  while ( !(stateflag & DVD_EXIT))
  {
		result = waitForCommand();
		switch (result)
		{
			case WMCMD_DVD_FWD:
fprintf(stdout, "dvdplayer:DEBUG Command = WMCMD_DVD_FWD\n");
				ProcessScan(&stateflag, DVD_SCAN_FORWARD);
				break;
			case WMCMD_DVD_PLAY:
fprintf(stdout, "dvdplayer:DEBUG Command = WMCMD_DVD_PLAY\n");
				ProcessPlay(&stateflag);
				break;
			case WMCMD_DVD_PAUSE:
fprintf(stdout, "dvdplayer:DEBUG Command = WMCMD_DVD_PAUSE\n");
				ProcessPause(&stateflag);
				break;
			case WMCMD_DVD_REV:
fprintf(stdout, "dvdplayer:DEBUG Command = WMCMD_DVD_REV\n");
				ProcessScan(&stateflag, DVD_SCAN_REVERSE);
				break;
			case WMCMD_DVD_STOP:
fprintf(stdout,"dvdplayer:DEBUG Command = WMCMD_DVD_STOP\n");
				ProcessStop(&stateflag);
				if (!(stateflag & DVD_ACTION_MASK))
					stateflag |= DVD_EXIT;
		} // end of switch
  } // end of while

	// Do house cleaning (i.e. shutdown the api)
fprintf(stdout, "dvdplayer:DEBUG Closing API and exiting wrapper application\n");
	VMW_DVD_Close();

  return(0);
}

