/**************************************************************************
   mpegplayer.c -- Plays MPEG transport stream, System and PES files ***
			using Sigma designs 8400 decoder ******

   Author: Muhammad Kafil, National Semiconductor, Santa Clara

   Modified for http streaming: Jeff Moellmer, Century Embedded Technologies
                                jeffm@censoft.com

   Modified for single instantiation within WebMedia: Jeff Webb, Century Embedded Tech.
								jeffw@censoft.com
   Modified for using mpeg middleware instead of using direct FMP Calls:
			Chitra Shankaran

**************************************************************/



/*
**
** Imported "Include" Files
**
*/
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <iostream.h>
#include <mpeg_api.h>

#define _DEFINE_FMP_TYPES_
#include "fmp.h"
#include "streamsrc_linux_http.hh"
#include "sharedclient.h"



/*
**
** Local Constant Defintions
**
*/
#define TRUE 1
#define FALSE 0



/*
**
** Local Variable Declarations
**
*/
StreamSource_HTTP *http;
static int data_source;
static int fd;
const char			*lckfile = "/var/lock/mpeg.lock";			/* Default lock file name */


/*
**
** Static Function prototypes
**
*/
static int InstanceRunning(void);
static int UpdateLockFile(const char *lockfile, pid_t pid);


/*
**
**
**
*/
static int openDataSource(void)
{
	// local file data source
	if(data_source == 0)
	{
		fd = open(cmd_block->parm1,O_RDONLY);
		if(fd == -1)
			return(-1);
	}

	// streaming webserver data source
	else if(data_source == 1)
	{
		http = new StreamSource_HTTP();
		if(http == NULL)
			return(-1);
		http->Init(cmd_block->parm1);
	}

	// exit with no errors
	return(0);
}



/*
**
**
**
*/
static int readDataSource(unsigned char *buf, int bufsize)
{
	int total;

	// local file data source
	if(data_source == 0)
	{
		while(1)	
		{
			total = read(fd,buf,bufsize);
			if((total == -1) && (errno == EINTR))
				continue;
			if(total == -1)
				return(-1);
			break;
		}
	}
	// streaming webserver data source
	else if(data_source == 1)
	{
		while(1)
		{
			total = http->FillBuffer(buf,bufsize);
			if((total == -1) && (errno == EINTR))
				continue;
			if(total == -1)
				return(-1);
			break;
		}
	}

	// exit with no errors
	return(total);
}



/*
**
**
**
*/
static int closeDataSource(void)
{
	// local file data source
	if(data_source == 0)
	{
		close(fd);
	}

	// streaming webserver data source
	else if(data_source == 1)
	{
		delete http;
	}

	// exit with no errors
	return(0);
}



/*
**
**
**
*/
static int waitForCommand(void)
{
	int cmd;

	while(1)
	{
		usleep(100000);
		if(cmd_block->status != WMSTATUS_START)
			continue;

		if((cmd_block->command == WMCMD_MPEG_STOP) ||
		   (cmd_block->command == WMCMD_MPEG_PLAY) ||
		   (cmd_block->command == WMCMD_MPEG_PAUSE))
			break;
	}

	cmd = cmd_block->command;
	cmd_block->status = WMSTATUS_DONE;
	while(cmd_block->status != WMSTATUS_IDLE)
		;

	return(cmd);
}



/*
**
**
**
*/
static int checkForCommand(void)
{
	int cmd;
	cmd = WMCMD_NONE;

	if (cmd_block->status == WMSTATUS_START)
	{
		if ((cmd_block->command == WMCMD_MPEG_STOP) ||
		    (cmd_block->command == WMCMD_MPEG_PLAY) ||
		    (cmd_block->command == WMCMD_MPEG_PAUSE))
		{
			cmd = cmd_block->command;
			cmd_block->status = WMSTATUS_DONE;
			while (cmd_block->status != WMSTATUS_IDLE);
		} // end of if
	} // end of if

	return (cmd);
}


/*******************************************************************************\
**
**	Function:	int InstanceRunning()
**	Desc:		Determines if an instance is already running
**	Accepts:	Nothing (void)
**	Returns:	int; 1 on instance running, 0 otherwise
**
\*******************************************************************************/
static int
InstanceRunning(void)
{
	int	retval = 0; /* Default return value */
	pid_t	mypid = getpid(), /* This processes pid */
			lockpid;  /* Pid of process from lock file */
	FILE	*lockfp; /* Lock file stream pointer */

	/* Open the file */
	if ( (lockfp = fopen(lckfile, "r")) == NULL)
	{
		/* There is no process running, add me to the file */
		retval = UpdateLockFile(lckfile, mypid);
		return (retval);
	} /* end of if */

	/*
	** A file existed, read the pid from the file
	*/
	if (fscanf(lockfp, "%d", &lockpid) != 1)
	{
		/* Unable to get the data from the file? */
		fclose (lockfp);
		retval = UpdateLockFile(lckfile, mypid);
		return (retval);
	} /* end of if */

	/*
	** See if this process is still running
	*/
	if (kill(lockpid, 0) == -1 && errno == ESRCH)
	{
		/* Process does not exist */
		fclose (lockfp);
		retval = UpdateLockFile(lckfile, mypid);
		return (retval);
	} /* end of if */

	fclose(lockfp);
	return (lockpid);
} /* end of InstanceRunning() */

/*******************************************************************************\
**
**	Function:	int UpdateLockFile()
**	Desc:		Creates the mpeg lock file adding this pid to the file
**	Accepts:	const char *lockfile = Ptr to the full path of the lock file
**				pid_t pid = Pid to add to the lock file
**	Returns:	int; 0 on success, -1 on error
**
\*******************************************************************************/
static int
UpdateLockFile(const char *lockfile, pid_t pid)
{
	FILE	*lockfp; /* File pointer to the lock file */

	/* Attempt to open the file for writing */
	if ( (lockfp = fopen(lockfile, "w")) == NULL)
		return (-1);

	fprintf(lockfp, "%d", pid);
	fclose(lockfp);
	return (0);
} /* end of UpdateLockFile() */

/*
**
**
**
*/
int main(int argc, char *argv[])
{
	DWORD pts,stc;
	//FMP_BUFFER FMPBuf;
	DECODE_BUFFER_INFO FMPBuf;
	int result,source,stopflag,streamtype,total;
	char* tvstd;
	unsigned long PauseOn=0;

	// Need some mechanism to determine if an instance is already running
	// TODO:  This may need to be changed into a mutex/semaphore depending on
	// the requirements coming down from NSC or any of its vendors
	if (result = InstanceRunning())
	{
		/* There already is an instance running, just exit this instance.... */
		if (result == -1)
			fprintf(stderr, "Process terminating, unable to update 
				lock file.\n");
		else
			fprintf(stderr, "Process terminating, instance (%d) already 
				running.\n", result);

		exit (0);
	} /* end of if */

	// recover the shared memory cmd_block pointer
	if(clientMemoryAttach() != 0)
	{
		fprintf(stderr,"Unable to attach to shared memory...\n");
		exit(1);
 	}

	// open MPEG driver
	streamtype = FMPF_SYSTEM;
	VMW_DVM_Open(streamtype,32*1024,8,NULL);
	tvstd = getenv("TVSTD");
	
	if(strcmp(tvstd,"pal") == 0)
	{
		printf("THe tvstd is set in %s \n",tvstd);
		FMPSet(FMPI_VIDEOOUT,FMPV_VIDEOOUT_PAL);
 	}
	else
		FMPSet(FMPI_VIDEOOUT,FMPV_VIDEOOUT_NTSC);
  	printf("Set the environment variable \n");
	// forever loop - never exits
	while(1)
 	{
		data_source = 0;
		if(strncmp(cmd_block->parm1,"http://",7) == 0)
		      data_source = 1;

		// open the data source
		result = openDataSource();
		if(result != 0)
			continue;
      
		// play this MPEG
		VMW_DVM_Play();
		stopflag = 0;
		total = 1;
		while(total)
   		{
			// check for a PAUSE or STOP command
			result = checkForCommand();
			if(result != WMCMD_NONE)
			{
				// pause command
				if(result == WMCMD_MPEG_PAUSE)
				{
					if(!PauseOn)
						PauseOn = 1;
					else if(PauseOn)
						PauseOn = 0;
					VMW_DVM_Pause(PauseOn);	
					while(1)
	  				{
						result = checkForCommand();
						if((result == WMCMD_MPEG_PLAY) ||
						   (result == WMCMD_MPEG_PAUSE))
	      						break;
	  				}
					PauseOn = 0;
					VMW_DVM_Play();
				}

				// stop command
				if(result == WMCMD_MPEG_STOP)
				{
					fprintf(stderr, "Received a WMCMD_MPEG_STOP 
						command.\n");
					stopflag = 1;
					total = 0;
					//continue;
						break;
				}
      			}	

		        // recover a buffer from the driver
			result = VMW_DVM_GetBufferPointer(&FMPBuf);
			if(result != VMW_SUCCESS)
     			{
				fprintf(stderr,"FATAL: unable to get an FMPBuffer...\n");
				return(-1);
			}

			// recover the next block of MPEG data
			total = readDataSource(FMPBuf.pBuffer,FMPBuf.bufferSize);
			if(total < 0)
			{
				fprintf(stderr,"FATAL: unable to read MPEG data...\n");
				return(-2);
			}
			FMPBuf.dataSize = total;

			// que the buffer in the driver
			result = VMW_DVM_StartMPEGDecode((U32*)&FMPBuf);
			if(result != VMW_SUCCESS)
     			{
				fprintf(stderr,"FATAL: unable to start decoding...\n");
				return(-1);
			}
   		 }

		// wait for playback completion
		if(stopflag == 0)
   		{
			while(1)
			{
				usleep(100000);
				pts = FMPGet(FMPI_PTS);
				stc = FMPGet(FMPI_STC);
				if(stc >= pts)
					exit(0);
			}
   		 }

		// do housekeeping
		VMW_DVM_Stop();
		closeDataSource();

		if (stopflag)
			break;
  	}

	// do housekeeping
	result = unlink(lckfile);
	if (result)
		fprintf(stderr, "unlink() failed on %s, error=%s (%d).\n", 
			lckfile, strerror(errno), errno);
	VMW_DVM_Close();

	// we should never get here
	return(0);
}



#ifdef NEVER

typedef int BOOL;
#define TRUE 1
#define FALSE 0
#define _DEFINE_FMP_TYPES_
#include "fmp.h"

#include "streamsrc_linux_http.hh"
#include "sharedmem.h"

void dump2(BYTE *mem, int size)
{
	for ( int i=0; i < size; i++) 
	cerr << hex << (int)mem[i] << " ";
}

int main(int argc,char **argv)
{
	int fd;
	int streamtype;
	int bytesread=1,rc=VMW_SUCCESS;

	// recover the shared memory cmd_block pointer
	if(sharedMemoryAttach() != 0)
	{
		fprintf(stderr,"Unable to attach to shared memory...\n");
		exit(1);
	}

	while(1)
	{
		// wait for an MPEG play command
		while(1)
  		{
			if((cmd_block->status == WMSTATUS_START) &&
			(cmd_block->command == WMCMD_MPEG_PLAY))
				break;
			usleep(100000);
		}
		cmd_block->status = WMSTATUS_DONE;

		// set 's' mode ( whatever the heck that is)
		streamtype=FMPF_SYSTEM;
		VMW_DVM_Open(streamtype,32*1024,8,NULL);
	
		VMW_DVM_Play();

		if ( strstr(cmd_block->parm1, "http://" ))
   		{
			StreamSource_HTTP *http = new StreamSource_HTTP();
			http->Init(cmd_block->parm1);
	  
			while (rc==VMW_SUCCESS)
			{
				while(bytesread)
				{
					FMP_BUFFER FMPBuf;
					rc = VMW_DVM_GetBufferPointer(&FMPBuf);
	
					if (rc != VMW_SUCCESS)
					{
						//printf("FMPError : %s\n", 
							FMPGetLastError(rc));
	  						break;
					}

					bytesread=http->FillBuffer(FMPBuf.pBuffer,
							FMPBuf.bufferSize);
	  
					if ( bytesread )
					{
						FMPBuf.dwDataSize=bytesread;

						// the buffer is read to be queued
						VMW_DVM_StartMPEGDecode((U32*)&FMPBuf);
					}
				}

				DWORD pts, stc;
				pts = FMPGet(FMPI_PTS);
				stc = FMPGet(FMPI_STC);
	    
				while (stc < pts)
				{
					pts = FMPGet(FMPI_PTS);
					stc = FMPGet(FMPI_STC);
	      
					if (stc >= pts)
					{
						VMW_DVM_Stop();
						VMW_DVM_Close();
						exit(0);
					}
				}
	    
				// Just in case
				VMW_DVM_Stop();
				VMW_DVM_Close();
				exit(0);
		      }
    		}
		else
		{
	        	fd=open(cmd_block->parm1,O_RDONLY);
			if (fd==-1)
			{
				perror("could not open file to play"); 
				exit(1); 
    			}
	  
			while (rc==VMW_SUCCESS)
			{
				while (bytesread)
				{
					FMP_BUFFER FMPBuf;

					// get a pointer to a free buffer
					rc = VMW_DVM_GetBufferPointer(&FMPBuf);
					if (rc != VMW_SUCCESS)
					{
						//printf ("FMPError : %s\n", 
							FMPGetLastError(rc));
							break;
					}	
	      
					// signals can cause harmless EINTR
					while (((bytesread=read(fd,FMPBuf.pBuffer,FMPBuf.bufferSize))==-1)
		 &&(errno==EINTR));

					if (bytesread==-1) 
						bytesread=0;
	      
					FMPBuf.dwDataSize=bytesread;
	      
					// the buffer is ready to be queued
					VMW_DVM_StartMPEGDecode(&FMPBuf);
				}
	    
				DWORD pts, stc;
				pts = FMPGet(FMPI_PTS);
				stc = FMPGet(FMPI_STC);

				while ( stc < pts )
				{
					pts = FMPGet(FMPI_PTS);
					stc = FMPGet(FMPI_STC);
	    
					if (stc >= pts)
					{
						close(fd);
						VMW_DVM_Stop();
						VMW_DVM_Close();
						exit(0);
	 				}
				}
	  
				close(fd);
				VMW_DVM_Stop();
				VMW_DVM_Close();
				exit(0);
			}
		}
	}
	exit(-1);
}

#endif
