/*
**
** Imported "Include" Files
**
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rsctools.h"



/*
**
** Constant Declarations
**
*/
#define MAX_WORKBUF_SIZE 256



/*
**
** Structure Definitions
**
*/
typedef enum
{
  PARSEMODE_DELIMITER,
  PARSEMODE_NEWLINE,
  PARSEMODE_TOKEN,
  PARSEMODE_WHITESPACE,
  PARSEMODE_UNDEFINED,
  PARSEMODE_LAST
}ParseMode;



/*
**
** Local Variable Definitions
**
*/
static char workbuf[MAX_WORKBUF_SIZE];
static unsigned short workbuf_index;
static char comment_list[] = {'!','*','|','#',0x00};
static char delimiter_list[] = {',',';',0x00};
static char whitespace_list[] = {0x09,' ',0x00};



/*
**
** This function will scan the specified list of characters for a match to
** the specified character. If the specified character is found in the
** list, a non-zero value will be returned. If not found, zero is returned.
**
*/
static int scanList(char *list, char chr)
{
  unsigned int index;

  index = 0;
  while(*(list + index))
  {
    if(*(list + index++) == chr)
      return(1);
  }
  return(0);
}



/*
**
** This function will extract the next token from the resource file. If
** found, the token is copied to the buffer pointed to by the "result"
** parameter. If the token is not found, the buffer will be empty.
**
*/
static void getToken(FILE *rfp, char *result)
{
  ParseMode mode;

  mode = PARSEMODE_UNDEFINED;
  while(1)
  {
    switch(mode)
    {
      case PARSEMODE_DELIMITER:
        ++workbuf_index;
        mode = PARSEMODE_UNDEFINED;
        break;

      case PARSEMODE_NEWLINE:
        while(1)
        {
          /* recover the next line */
          if(fgets(workbuf,MAX_WORKBUF_SIZE,rfp) == NULL)
          {
            *result = 0x00;
            return;
          }

          /* remove the newline at the end of string */
          workbuf[strlen(workbuf) - 1] = 0x00;

          /* skip blank lines */
          if(workbuf[0] == 0x00)
            continue;

          /* skip comment lines */
          if(scanList(comment_list,workbuf[0]))
            continue;

          /* indicate new line loaded */
          workbuf_index = 0;
          mode = PARSEMODE_UNDEFINED;
          break;
        }
        break;

      case PARSEMODE_TOKEN:
        /* check for beginning of string - special case */
        if(workbuf[workbuf_index] == '"')
        {
          while(1)
          {
            /* check for end of string */
            ++workbuf_index;
            if(workbuf[workbuf_index] == '"')
            {
              ++workbuf_index;
              *result = 0x00;
              return;
            }

            /* check for end-of-line */
            if(workbuf[workbuf_index] == 0x00)
            {
              *result = 0x00;
              return;
            }

            /* check for escape sequence */
            if(workbuf[workbuf_index] == '\\')
            {
              ++workbuf_index;
              *result++ = workbuf[workbuf_index];
            }

            /* must be part of the token */
            else
              *result++ = workbuf[workbuf_index];
          }
        }

        /* process standard token */
        while(1)
        {
          /* check for end-of-line */
          if(workbuf[workbuf_index] == 0x00)
          {
            *result = 0x00;
            return;
          }

          /* check for end of token */
          if((scanList(comment_list,workbuf[workbuf_index])) ||
             (scanList(delimiter_list,workbuf[workbuf_index])) ||
             (scanList(whitespace_list,workbuf[workbuf_index])))
          {
            *result = 0x00;
            return;
          }

          /* must be part of the token */
          *result++ = workbuf[workbuf_index++];
        }
        break;

      case PARSEMODE_WHITESPACE:
        while(1)
        {
          /* check for end-of-line */
          if(workbuf[workbuf_index] == 0x00)
          {
            mode = PARSEMODE_NEWLINE;
            break;
          }

          /* check for whitespace */
          if((scanList(delimiter_list,workbuf[workbuf_index])) ||
                  (scanList(whitespace_list,workbuf[workbuf_index])))
          {
            ++workbuf_index;
          }

          /* check for comment */
          else if(scanList(comment_list,workbuf[workbuf_index]))
          {
            mode = PARSEMODE_NEWLINE;
            break;
          }

          /* must be the beginning of a token */
          else
          {
            mode = PARSEMODE_TOKEN;
            break;
          }
        }
        break;

      /* if unknown parsing mode, determine it */
      case PARSEMODE_UNDEFINED:
        if(scanList(delimiter_list,workbuf[workbuf_index]))
        {
          mode = PARSEMODE_DELIMITER;
          break;
        }
        if(workbuf[workbuf_index] == 0x00)
        {
          mode = PARSEMODE_NEWLINE;
          break;
        }
        if(scanList(whitespace_list,workbuf[workbuf_index]))
        {
          mode = PARSEMODE_WHITESPACE;
          break;
        }
        mode = PARSEMODE_TOKEN;
        break;
    }
  }
}



/*
**
** This function will search the specified file for the specified
** keyword. The search begins at the current file position, and if
** the keyword is not found when the end of the file is reached,
** the search will continue at the beginning of the file.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "keyword"
** parameter points to the buffer that is loaded with the keyword.
**
** If the keyword is found, '1' is returned and the file's position
** will be set to the first character following the keyword. If the
** keyword is not found, '0' is returned. If an error occurs during
** function execution, '-1' is returned.
**
*/
int locateKeyword(FILE *rfp, char *keyword)
{
  char buf[256];
  int index;
  unsigned int klen,pass;

  /* recover the string lengths */
  sprintf(buf,"%s:",keyword);
  klen = strlen(buf);

  /* scan the file for the keyword */
  pass = 0;
  while(pass < 2)
  {
    if(fgets(workbuf,MAX_WORKBUF_SIZE,rfp) == NULL)
    {
      rewind(rfp);
      ++pass;
    }
    else
    {
      /* remove the newline at the end of string */
      workbuf[strlen(workbuf) - 1] = 0x00;

      /* skip blank lines */
      if(workbuf[0] == 0x00)
        continue;

      /* skip comment lines */
      if(scanList(comment_list,workbuf[0]))
        continue;

      /* line must be equal or longer than the keyword */
      index = strlen(workbuf) - klen;
      if(index < 0)
        continue;

      /* valid text - check for keyword */
      do
      {
        if(strncmp(&workbuf[index],buf,klen) == 0)
        {
          workbuf_index = index + klen;
          return(1);
        }
      }while(--index >= 0);
    }
  }

  /* keyword not found - exit with errors */
  return(0);
}



/*
**
** This function will recover the next keyword entry found in the
** specified file. The search begins at the current file position,
** and terminates when the end of the file is reached.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "result"
** parameter points to a buffer where the next keyword found in the
** file will be stored.
**
** If successful, '1' is returned and the result buffer will be
** loaded with the next keyword. If no keyword is found, '0' is
** returned. If an error occurs during function execution, '-1' is
** returned. In any case, the file's position is not altered.
**
*/
int previewNextKeyword(FILE *rfp, char *result)
{
  char buf[256],wbuf[256];
  long filepos;
  unsigned short wbindex;

  /* recover the file's current position */
  filepos = ftell(rfp);
  strcpy(wbuf,workbuf);
  wbindex = workbuf_index;

  /* locate the next keyword */
  while(1)
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      strcpy(workbuf,wbuf);
      workbuf_index = wbindex;
      *result = 0x00;
      return(0);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
      break;
  }

  /* recover keyword */
  buf[strlen(buf) - 1] = 0x00;
  strcpy(result,buf);

  /* restore original file position */
  fseek(rfp,filepos,0);
  strcpy(workbuf,wbuf);
  workbuf_index = wbindex;

  /* exit with no errors */
  return(1);
}



/*
**
** This function will extract the number of parameters associated
** with a keyword. After the keyword has been located, the number
** of parameters that follow the keyword are calculated.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened.
**
** This function returns the number of parameters associated with the
** keyword. If an error occurs during function execution, '-1' is
** returned.
**
*/
int getParmTotal(FILE *rfp)
{
  char buf[256],wbuf[256];
  int tparm;
  long filepos;
  unsigned short wbindex;

  /* recover the current file position */
  filepos = ftell(rfp);
  strcpy(wbuf,workbuf);
  wbindex = workbuf_index;

  /* recover all requested parameters */
  tparm = 0;
  while(1)
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
      break;

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
      break;

    /* must be valid parameter - bump the total */
    ++tparm;
  }

  /* restore file position */
  fseek(rfp,filepos,0);
  strcpy(workbuf,wbuf);
  workbuf_index = wbindex;

  /* return result and exit with no errors */
  return(tparm);
}



/*
**
** This function will recover the byte values associated with a
** keyword. After the keyword is located, the parameters that follow
** are recovered as signed byte values.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "result"
** parameter points to an array of bytes where each byte value will
** be stored. The "count" parameter contains the number of byte
** parameters that will be recovered.
**
** This function returns the number of byte parameters recovered.
** If an error occurs during function execution, '-1' is returned.
**
** If the "count" parameter is zero, then this function will recover
** any and all byte parameters associated with the keyword. This
** process will continue until the next keyword is found or the
** end-of-file is detected.
**
*/
int getByteParm(FILE *rfp, char *result, unsigned int count)
{
  char buf[256];
  long filepos;
  unsigned int tparm;

  /* recover the file's current position */
  filepos = ftell(rfp);

  /* recover all requested parameters */
  tparm = 0;
  do
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* convert parameter to byte value and store it */
    *result++ = (char)atoi(buf);
    ++tparm;
  }while(--count);

  /* exit with no errors */
  fseek(rfp,filepos,0);
  return(tparm);
}



/*
**
** This function will recover the integer values associated with a
** keyword. After the keyword is located, the parameters that follow
** are recovered as signed integer values.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "result"
** parameter points to an array of integers where each integer value
** will be stored. The "count" parameter contains the number of
** integer parameters that will be recovered.
**
** This function returns the number of integer parameters recovered.
** If an error occurs during function execution, '-1' is returned.
**
** If the "count" parameter is zero, then this function will recover
** any and all integer parameters associated with the keyword. This
** process will continue until the next keyword is found or the
** end-of-file is detected.
**
*/
int getIntegerParm(FILE *rfp, int *result, unsigned int count)
{
  char buf[256];
  long filepos;
  unsigned int tparm;

  /* recover the file's current position */
  filepos = ftell(rfp);

  /* recover all requested parameters */
  tparm = 0;
  do
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* convert parameter to integer value and store it */
    *result++ = (int)atoi(buf);
    ++tparm;
  }while(--count);

  /* exit with no errors */
  fseek(rfp,filepos,0);
  return(tparm);
}



/*
**
** This function will recover the double values associated with a
** keyword. After the keyword is located, the parameters that follow
** are recovered as floating-point, "double" values.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "result"
** parameter points to an array of doubles where each double value
** will be stored. The "count" parameter contains the number of
** double parameters that will be recovered.
**
** This function returns the number of double parameters recovered.
** If an error occurs during function execution, '-1' is returned.
**
** If the "count" parameter is zero, then this function will recover
** any and all "double" parameters associated with the keyword. This
** process will continue until the next keyword is found or the
** end-of-file is detected.
**
*/
int getDoubleParm(FILE *rfp, double *result, unsigned int count)
{
  char buf[256];
  long filepos;
  unsigned int tparm;

  /* recover the file's current position */
  filepos = ftell(rfp);

  /* recover all requested parameters */
  tparm = 0;
  do
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* convert parameter to byte value and store it */
    *result++ = (double)atof(buf);
    ++tparm;
  }while(--count);

  /* exit with no errors */
  fseek(rfp,filepos,0);
  return(tparm);
}



/*
**
** This function will recover the text strings associated with a
** keyword. After the keyword is located, the parameters that follow
** are recovered as text strings.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "result"
** parameter points to an array of string pointers. Each string
** pointer points to a buffer where one of the text strings is to be
** stored. The "count" parameter contains the number of text strings
** that will be recovered.
**
** This function returns the number of text strings recovered.
** If an error occurs during function execution, '-1' is returned.
**
** If the "count" parameter is zero, then this function will recover
** any and all text strings associated with the keyword. This process
** will continue until the next keyword is found or the end-of-file
** is detected.
**
*/
int getStringParm(FILE *rfp, char **result, unsigned int count)
{
  char buf[256];
  long filepos;
  unsigned int tparm;

  /* recover the file's current position */
  filepos = ftell(rfp);

  /* recover all requested parameters */
  tparm = 0;
  do
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* copy the string parameter */
    strcpy(*result++,buf);
    ++tparm;

  }while(--count);

  /* exit with no errors */
  fseek(rfp,filepos,0);
  return(tparm);
}



/*
**
** This function will recover the parameters associated with a
** keyword. The parameters that are recovered may be any combination
** of byte, integer, double, or text strings.
**
** The "rfp" parameter points to the FILE structure that was returned
** when the resource file was successfully opened. The "parmlist"
** parameter points to an array of ParmEntry structures, where each
** structure is initialized with the expected parameter type and a
** pointer to the location where the resulting data will be stored.
** The "count" parameter contains the number of parameters that will
** be recovered from the resource file.
**
** This function returns the number of parameters recovered. If an
** error occurs during function execution, '-1' is returned.
**
*/
int getMixedParm(FILE *rfp, ParmEntry *parmlist, unsigned int count)
{
  char buf[256];
  long filepos;
  unsigned int tparm;

  /* recover the file's current position */
  filepos = ftell(rfp);

  /* recover all requested parameters */
  tparm = 0;
  do
  {
    /* recover a token */
    getToken(rfp,buf);

    /* check for valid token */
    if(buf[0] == 0x00)
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* check for new keyword */
    if(buf[strlen(buf) - 1] == ':')
    {
      fseek(rfp,filepos,0);
      return(tparm);
    }

    /* recover the parameter data */
    switch(parmlist->type)
    {
      case PARMTYPE_BYTE:
        (*((char *)parmlist->result)) = ((char)atoi(buf));
        break;

      case PARMTYPE_INTEGER:
        (*((int *)parmlist->result)) = ((int)atoi(buf));
        break;

      case PARMTYPE_DOUBLE:
        (*((double *)parmlist->result)) = ((double)atof(buf));
        break;

      case PARMTYPE_STRING:
        strcpy(((char *)parmlist->result),buf);
        break;

      default:
        fseek(rfp,filepos,0);
        return(-1);
    }
    ++tparm;
    ++parmlist;
  }while(--count);

  /* exit with no errors */
  fseek(rfp,filepos,0);
  return(tparm);
}



/*
**
** This function will recover the raw text on the line associated
** with the keyword. The newline on the end of the line will be
** removed during the recovery.
**
** The "result" parameter points to the buffer that will hold the
** resulting text.
**
*/
void getRawString(char *result)
{
  while(workbuf[workbuf_index] != 0)
    *result++ = workbuf[workbuf_index++];
  *result = 0x00;
}
